/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrender
 * @short_description: FIXME
 * @see_also: FIXME
 *
 * FIXME
 */

#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <gst/gst.h>
#include "pgmrender.h"

static gboolean pgm_render_initialized = FALSE;
static GModule *module = NULL;
static PgmRenderPluginDesc *pgm_render_plugin_desc = NULL;

PGM_RENDER_EXPORT PgmRenderPluginCreateWindowFunc pgm_render_create_window;
PGM_RENDER_EXPORT PgmRenderPluginCreateSurfaceFunc pgm_render_create_surface;
PGM_RENDER_EXPORT PgmRenderPluginCreateFontFunc pgm_render_create_font;

/* FIXME: The function doesn't handle colon-separated list of paths. */
static GModule *
find_plugin (const gchar *plugin_name, const gchar *plugin_directory)
{
  GDir *dir;
  const gchar *dir_entry;
  gchar *filename;
  gint plugin_name_len;
  gint loaded_plugin_name_len;

  plugin_name_len = strlen (plugin_name);

  dir = g_dir_open (plugin_directory, 0, NULL);
  if (!dir)
    return NULL;

  while ((dir_entry = g_dir_read_name (dir)))
    {
      filename = g_strjoin ("/", plugin_directory, dir_entry, NULL);

      module = g_module_open (filename, G_MODULE_BIND_LAZY);
      
      if (!module)
        continue;

      if (!g_module_symbol (module, "plugin_desc",
			    (gpointer) &pgm_render_plugin_desc))
	{
	  g_print ("Can't load symbol \"plugin_desc\"\n");
	
	  if (!g_module_close (module))
            g_print ("Can't close module \"%s\"\n", filename);
	
	  exit (1);
	}

      loaded_plugin_name_len = strlen (pgm_render_plugin_desc->name);

      if (!g_ascii_strncasecmp (pgm_render_plugin_desc->name, plugin_name,
				MAX (plugin_name_len, loaded_plugin_name_len)))
	break;
      else
	{
	  g_module_close (module);
	  module = NULL;
	}

      g_free (filename);
    }

  g_dir_close (dir);

  return module;
}

/**
 * pgm_render_init:
 * @plugin_name: The name of the plugin to load
 *
 * FIXME
 */
void
pgm_render_init (const gchar *plugin_name)
{
  const gchar *plugin_directory;

  gst_init (NULL, NULL);

  plugin_directory = g_getenv (PLUGIN_PATH_VARIABLE_NAME);

  if (plugin_directory)
    module = find_plugin (plugin_name, plugin_directory);

  if (!module)
    module = find_plugin (plugin_name, PGM_PLUGIN_DIRECTORY);

  if (!module)
    {
      g_print ("Can't find any module with name '%s'\n", plugin_name);
      exit (1);
    }

  pgm_render_create_window = pgm_render_plugin_desc->create_window;
  pgm_render_create_surface = pgm_render_plugin_desc->create_surface;
  pgm_render_create_font = pgm_render_plugin_desc->create_font;

  /* FIXME: Add the GStreamer logging system and make it a debug message. */
/*  g_print ("DEBUG:\n" */
/* 	   "create_window func  : %p\n" */
/* 	   "create_surface func : %p\n" */
/* 	   "create_font func    : %p\n" */
/* 	   "Name                : %s\n" */
/* 	   "Description         : %s\n" */
/* 	   "License             : %s\n" */
/* 	   "Origin              : %s\n\n", */
/* 	   pgm_render_plugin_desc->create_window, */
/* 	   pgm_render_plugin_desc->create_surface, */
/* 	   pgm_render_plugin_desc->create_font, */
/* 	   pgm_render_plugin_desc->name, */
/* 	   pgm_render_plugin_desc->description, */
/* 	   pgm_render_plugin_desc->license, */
/* 	   pgm_render_plugin_desc->origin); */

  pgm_render_initialized = TRUE;
}

/**
 * pgm_render_deinit:
 *
 * FIXME
 */
void
pgm_render_deinit ()
{
  if (!g_module_close (module))
    g_print ("Can't close module\n");

  pgm_render_initialized = FALSE;
}
