/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrenderpipe3d-iface
 * @short_description: An interface to embed 3D API.
 *
 * #PgmRenderPipe3dIFace is an interface to add 3D drawing capabilities
 * to an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrenderpipe3d-iface.h"

static void pgm_render_pipe3d_iface_base_init (gpointer g_class);

static void
pgm_render_pipe3d_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_pipe3d_iface_get_type (void)
{
  static GType type = 0;

  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderPipe3dIFaceClass),
	pgm_render_pipe3d_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };      
      type = g_type_register_static (G_TYPE_INTERFACE,
				     "PgmRenderPipe3dIFace", &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_pipe3d_set_translation:
 * @pipe3d: A #PgmRenderPipe3dIFace object.
 * @x: The x coordinate of the translation.
 * @y: The y coordinate of the translation.
 * @z: The z coordinate of the translation.
 *
 * Add a translation to an object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_pipe3d_set_translation (PgmRenderPipe3dIFace *pipe3d,
				   gfloat x, gfloat y, gfloat z)
{
  g_return_val_if_fail (PGM_RENDER_IS_PIPE3D_IFACE (pipe3d),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_PIPE3D_IFACE_GET_CLASS
    (pipe3d)->set_translation (pipe3d, x, y, z);
}

/**
 * pgm_render_pipe3d_get_translation:
 * @pipe3d: A #PgmRenderPipe3dIFace object.
 * @x: The x coordinate of the translation to fill.
 * @y: The y coordinate of the translation to fill.
 * @z: The z coordinate of the translation to fill.
 *
 * Retrieves the translation of an object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_pipe3d_get_translation (PgmRenderPipe3dIFace *pipe3d,
				   gfloat *x, gfloat *y, gfloat *z)
{
  g_return_val_if_fail (PGM_RENDER_IS_PIPE3D_IFACE (pipe3d),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_PIPE3D_IFACE_GET_CLASS
    (pipe3d)->get_translation (pipe3d, x, y, z);
}

/**
 * pgm_render_pipe3d_set_rotation:
 * @pipe3d: A #PgmRenderPipe3dIFace object.
 * @x: The x angle of rotation in degrees.
 * @y: The y angle of rotation in degrees.
 * @z: The z angle of rotation in degrees.
 *
 * Add a rotation in degrees to an object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_pipe3d_set_rotation (PgmRenderPipe3dIFace *pipe3d, gfloat x,
				gfloat y, gfloat z)
{
  g_return_val_if_fail (PGM_RENDER_IS_PIPE3D_IFACE (pipe3d),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_PIPE3D_IFACE_GET_CLASS
    (pipe3d)->set_rotation (pipe3d, x, y ,z);
}

/**
 * pgm_render_pipe3d_get_rotation:
 * @pipe3d: A #PgmRenderPipe3dIFace object.
 * @x: The x angle of rotation in degrees to fill.
 * @y: The y angle of rotation in degrees to fill.
 * @z: The z angle of rotation in degrees to fill.
 *
 * Retrieves the rotation in degrees of an object.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_pipe3d_get_rotation (PgmRenderPipe3dIFace *pipe3d, gfloat *x,
				gfloat *y, gfloat *z)
{
  g_return_val_if_fail (PGM_RENDER_IS_PIPE3D_IFACE (pipe3d),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_PIPE3D_IFACE_GET_CLASS
    (pipe3d)->get_rotation (pipe3d, x, y, z);
}
