/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render OpenGL 1.2 plugin
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 *
 * Author: Loïc Molinari <loic@fluendo.com>
 */

#include "pgmrendergl1font.h"
#include "pgmrendergl1window.h"
#include <string.h>

static void pgm_render_gl1_font_class_init (PgmRenderGl1FontClass *klass);
static void pgm_render_gl1_font_init (PgmRenderGl1Font *obj);
static void pgm_render_gl1_font_finalize (GObject *obj);

static void layout_iface_init (PgmRenderLayoutIFaceClass *iface);
static void text_iface_init (PgmRenderTextIFaceClass *iface);
static void pipe3d_iface_init (PgmRenderPipe3dIFaceClass *iface);

static PgmRenderError pgm_render_gl1_font_draw (PgmRenderDrawable *drawable);
static gint get_upper_power_of_two (gint number);
static void generate_cairo_label (PgmRenderGl1Font *obj);
static void set_size (PgmRenderGl1Font *gl1_font);

static PgmRenderDrawableClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE (PgmRenderGl1Font, pgm_render_gl1_font,
			 PGM_RENDER_TYPE_DRAWABLE,
			 {
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_LAYOUT_IFACE, layout_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_TEXT_IFACE, text_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_PIPE3D_IFACE, pipe3d_iface_init);
			 });

static void
pgm_render_gl1_font_class_init (PgmRenderGl1FontClass *klass)
{
  GObjectClass *gobject_class;
  PgmRenderDrawableClass *drawable_class;

  gobject_class = (GObjectClass*) klass;
  drawable_class = (PgmRenderDrawableClass*) klass;

  drawable_class->draw = pgm_render_gl1_font_draw;
  parent_class = g_type_class_peek_parent (klass);
  gobject_class->finalize = pgm_render_gl1_font_finalize;

  klass->backend = NULL;
  klass->font_with_media_data = NULL;
}

static void
pgm_render_gl1_font_init (PgmRenderGl1Font *obj)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);
  /* PgmRenderBaseWindow *win = PGM_RENDER_DRAWABLE (gl1_font)->window; */
  gint i;

  gl1_font->text_font_name = g_strdup ("Sans");
  gl1_font->text_label = g_strdup (" ");
  gl1_font->text_size = 16;
  gl1_font->text_slant = CAIRO_FONT_SLANT_NORMAL;
  gl1_font->text_weight = CAIRO_FONT_WEIGHT_BOLD;

  /* Set font options */
  gl1_font->text_font_options = cairo_font_options_create ();
  cairo_font_options_set_hint_metrics (gl1_font->text_font_options,
				       CAIRO_HINT_METRICS_DEFAULT);
  cairo_font_options_set_hint_style (gl1_font->text_font_options,
				     CAIRO_HINT_STYLE_DEFAULT);
  cairo_font_options_set_subpixel_order (gl1_font->text_font_options,
					 CAIRO_SUBPIXEL_ORDER_DEFAULT);
  cairo_font_options_set_antialias (gl1_font->text_font_options,
				    CAIRO_ANTIALIAS_DEFAULT);

  for (i = 0; i < 4; i++)
    {
      obj->text_vertex[i].x = 0.0f;
      obj->text_vertex[i].y = 0.0f;
      obj->text_vertex[i].z = 0.0f;
      obj->text_vertex[i].w = 1.0f;
    }

  obj->text_coord[0].s = 0.0f; obj->text_coord[0].t = 0.0f;
  obj->text_coord[0].r = 0.0f; obj->text_coord[0].q = 1.0f;
  obj->text_coord[1].s = 0.0f; obj->text_coord[1].t = 1.0f;
  obj->text_coord[1].r = 0.0f; obj->text_coord[1].q = 1.0f;
  obj->text_coord[2].s = 1.0f; obj->text_coord[2].t = 1.0f;
  obj->text_coord[2].r = 0.0f; obj->text_coord[2].q = 1.0f;
  obj->text_coord[3].s = 1.0f; obj->text_coord[3].t = 0.0f;
  obj->text_coord[3].r = 0.0f; obj->text_coord[3].q = 1.0f;

  obj->layout_width = 200;
  obj->layout_height = 50;
  obj->layout_x = 0.0f;
  obj->layout_y = 0.0f;
  obj->layout_x_align = 0.5f;
  obj->layout_y_align = 0.5f;

  obj->text_color.r = obj->text_color.g = obj->text_color.b =
    obj->text_color.a = 1.0f;

/*   if (win) */
/*     { */
/*       obj->sx = win->sx; */
/*       obj->sy = win->sy; */
/*       obj->window_sx = win->sx; */
/*       obj->window_sy = win->sy; */
/*     } */
/*   else */
/*     { */
  obj->sx = 1.0f;
  obj->sy = 1.0f;
  obj->window_sx = 1.0f;
  obj->window_sy = 1.0f;
/*     } */
  /* generate_cairo_label (gl1_font); */
}

static void
pgm_render_gl1_font_finalize (GObject *obj)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_free (gl1_font->text_pixels);
  g_free (gl1_font->text_font_name);
  g_free (gl1_font->text_label);

  G_OBJECT_CLASS (parent_class)->finalize (obj);
}

PgmRenderGl1Font *
pgm_render_gl1_font_new (void)
{
  return PGM_RENDER_GL1_FONT (g_object_new (PGM_RENDER_TYPE_GL1_FONT, NULL));
}

/* Utility functions */

static gint
get_upper_power_of_two (gint number)
{
  gint power_of_two = 1;

  while (power_of_two < number)
    power_of_two <<= 1;

  return power_of_two;
}

static void
generate_cairo_label (PgmRenderGl1Font *obj)
{
  PgmRenderGl1FontClass *gl1_font_class = PGM_RENDER_GL1_FONT_GET_CLASS (obj);
  gdouble move_x, move_y;

  obj->text_surface_extents =
    cairo_image_surface_create (CAIRO_FORMAT_ARGB32, 0, 0);
  obj->text_cr_extents = cairo_create (obj->text_surface_extents);
  cairo_scale (obj->text_cr_extents, obj->window_sx, obj->window_sy);

  cairo_select_font_face (obj->text_cr_extents, obj->text_font_name,
			  obj->text_slant, obj->text_weight);
  cairo_set_font_size (obj->text_cr_extents, obj->text_size);
  cairo_font_extents (obj->text_cr_extents, &obj->text_font_extents);

  /* The Cairo Win32 Text backend doesn't seem to like empty strings */
  if (strlen (obj->text_label))
    cairo_text_extents (obj->text_cr_extents, obj->text_label,
                        &obj->text_text_extents);
  else
    cairo_text_extents (obj->text_cr_extents, " ", &obj->text_text_extents);

  obj->text_width = (obj->text_text_extents.width + 2) * obj->window_sx;
  obj->text_tex_width = get_upper_power_of_two (obj->text_width);
  obj->text_height = (obj->text_font_extents.ascent +
                      obj->text_font_extents.descent + 2) * obj->window_sy;
  obj->text_tex_height = get_upper_power_of_two (obj->text_height);

  /* Create the pixel buffer image */
  if (obj->text_pixels != NULL)
    g_free (obj->text_pixels);  
  obj->text_pixels = (guchar *) g_malloc (4 * (obj->text_tex_width) *
					  (obj->text_tex_height));
  obj->text_surface = cairo_image_surface_create_for_data
    (obj->text_pixels, CAIRO_FORMAT_ARGB32, obj->text_tex_width,
     obj->text_tex_height, obj->text_tex_width * 4);
  obj->text_cr = cairo_create (obj->text_surface);

  /* Set the font options */
  cairo_set_font_options (obj->text_cr, obj->text_font_options);

  /* Fill background with full alpha */
  cairo_rectangle (obj->text_cr, 0.0f, 0.0f, obj->text_tex_width,
		   obj->text_tex_height);
  cairo_set_source_rgba (obj->text_cr, 0.0f, 0.0f, 0.0f, 0.0f);
  cairo_set_operator (obj->text_cr, CAIRO_OPERATOR_SOURCE);
  cairo_fill (obj->text_cr);

  /* Draw the text in white */
  cairo_scale (obj->text_cr, obj->window_sx, obj->window_sy);
  cairo_select_font_face (obj->text_cr, obj->text_font_name,
			  obj->text_slant, obj->text_weight);
  cairo_set_font_size (obj->text_cr, obj->text_size);
  move_x = -obj->text_text_extents.x_bearing + 1;
  move_y = obj->text_font_extents.ascent + 1;
  cairo_set_source_rgba (obj->text_cr, 0.0f, 0.0f, 0.0f, 0.95f);
  cairo_move_to (obj->text_cr, move_x - 1, move_y - 1);
  cairo_show_text (obj->text_cr, obj->text_label);
  cairo_move_to (obj->text_cr, move_x + 1, move_y - 1);
  cairo_show_text (obj->text_cr, obj->text_label);
  cairo_move_to (obj->text_cr, move_x + 1, move_y + 1);
  cairo_show_text (obj->text_cr, obj->text_label);
  cairo_move_to (obj->text_cr, move_x - 1, move_y + 1);
  cairo_show_text (obj->text_cr, obj->text_label);
  cairo_set_source_rgba (obj->text_cr, 1.0f, 1.0f, 1.0f, 1.0f);
  cairo_move_to (obj->text_cr, move_x, move_y);
  cairo_show_text (obj->text_cr, obj->text_label);

  /* Limit the power-of-two texture to the size of the text, computing adapted
   * texture coordinates. */
  obj->text_coord_x_limit =
    (gfloat) obj->text_width / (gfloat) obj->text_tex_width;
  obj->text_coord_y_limit =
    (gfloat) (obj->text_height) / (gfloat) obj->text_tex_height;
  obj->text_coord[2].s = obj->text_coord[3].s = obj->text_coord_x_limit;
  obj->text_coord[1].t = obj->text_coord[2].t = obj->text_coord_y_limit;

  /* Update size */
  set_size (obj);

  /* Debugging facility */
/*   cairo_surface_write_to_png (obj->text_surface, */
/*                               "/home/loicm/Desktop/pgm-font.png"); */
  
  /* Add the font to the list of drawable to update on the next redraw */
  PGM_RENDER_DRAWABLE_LOCK (obj);
  gl1_font_class->font_with_media_data = g_list_prepend
    (gl1_font_class->font_with_media_data, obj);
  PGM_RENDER_DRAWABLE_UNLOCK (obj);
}

static void
set_size (PgmRenderGl1Font *gl1_font)
{
  gl1_font->text_coord[0].s = gl1_font->text_coord[1].s = 0.0f;
  gl1_font->text_coord[2].s = gl1_font->text_coord[3].s =
    gl1_font->text_coord_x_limit;
  gl1_font->text_coord[0].t = gl1_font->text_coord[3].t = 0.0f;
  gl1_font->text_coord[1].t = gl1_font->text_coord[2].t =
    gl1_font->text_coord_y_limit;

  gl1_font->text_vertex[0].x = gl1_font->text_vertex[1].x =
    gl1_font->layout_x * gl1_font->window_sx + gl1_font->layout_x_align
    * (gl1_font->layout_width * gl1_font->window_sx
       - gl1_font->text_width * gl1_font->sx);
  gl1_font->text_vertex[2].x = gl1_font->text_vertex[3].x =
    gl1_font->text_vertex[0].x + gl1_font->text_width * gl1_font->sx;

  gl1_font->text_vertex[0].y = gl1_font->text_vertex[3].y =
    gl1_font->layout_y * gl1_font->window_sy + gl1_font->layout_y_align
    * (gl1_font->layout_height * gl1_font->window_sy
       - gl1_font->text_height * gl1_font->sy);
  gl1_font->text_vertex[1].y = gl1_font->text_vertex[2].y =
    gl1_font->text_vertex[0].y + gl1_font->text_height * gl1_font->sy;
}

/* Layout interface implementation */

static PgmRenderError
gl1_font_ilayout_set_size (PgmRenderLayoutIFace *obj, gint width, gint height)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  gl1_font->layout_width = width;
  gl1_font->layout_height = height;

  set_size (gl1_font);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_get_size (PgmRenderLayoutIFace *obj, gint *width,
			   gint *height)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  *width = gl1_font->text_vertex[3].x - gl1_font->text_vertex[0].x;
  *height = gl1_font->text_vertex[1].y - gl1_font->text_vertex[0].y;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_set_position (PgmRenderLayoutIFace *obj, gint x, gint y)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  gl1_font->layout_x = x;
  gl1_font->layout_y = y;

  set_size (gl1_font);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_get_position (PgmRenderLayoutIFace *obj, gint *x, gint *y)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);

  *x = gl1_font->text_vertex[0].x;
  *y = gl1_font->text_vertex[0].y;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_set_alignment (PgmRenderLayoutIFace *obj, gfloat x,
				gfloat y)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  gl1_font->layout_x_align = x;
  gl1_font->layout_y_align = y;

  set_size (gl1_font);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_get_alignment (PgmRenderLayoutIFace *obj, gfloat *x,
				gfloat *y)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);

  *x = gl1_font->layout_x_align;
  *y = gl1_font->layout_y_align;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_set_aspect_style (PgmRenderLayoutIFace *obj,
				   PgmRenderLayoutAspectStyle style)
{
  g_print ("Gl1Font implementation of LayoutIFace set_aspect_style.\n");

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_get_aspect_style (PgmRenderLayoutIFace *obj,
				   PgmRenderLayoutAspectStyle *style)
{
  g_print ("Gl1Font implementation of LayoutIFace get_aspect_style.\n");

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_set_aspect_ratio (PgmRenderLayoutIFace *obj, gfloat ratio)
{
  g_print ("Gl1Font implementation of LayoutIFace set_aspect_ratio.\n");

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ilayout_get_aspect_ratio (PgmRenderLayoutIFace *obj, gfloat *ratio)
{
  g_print ("Gl1Font implementation of LayoutIFace get_aspect_ratio.\n");

  return PGM_RENDER_ERROR_X;
}

static void
layout_iface_init (PgmRenderLayoutIFaceClass *iface)
{
  iface->set_size = gl1_font_ilayout_set_size;
  iface->get_size = gl1_font_ilayout_get_size;
  iface->set_position = gl1_font_ilayout_set_position;
  iface->get_position = gl1_font_ilayout_get_position;
  iface->set_alignment = gl1_font_ilayout_set_alignment;
  iface->get_alignment = gl1_font_ilayout_get_alignment;
  iface->set_aspect_style = gl1_font_ilayout_set_aspect_style;
  iface->get_aspect_style = gl1_font_ilayout_get_aspect_style;
  iface->set_aspect_ratio = gl1_font_ilayout_set_aspect_ratio;
  iface->get_aspect_ratio = gl1_font_ilayout_get_aspect_ratio;
}

/* Text interface implementation */

static PgmRenderError
gl1_font_itext_set_label (PgmRenderTextIFace *text, const gchar *label)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  g_free (gl1_font->text_label);
  gl1_font->text_label = g_strndup (label, strlen (label));

  generate_cairo_label (gl1_font);

  return PGM_RENDER_ERROR_X;
}

/* FIXME */
static PgmRenderError
gl1_font_itext_get_label (PgmRenderTextIFace *text, gchar *label)
{
  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_itext_set_font (PgmRenderTextIFace *text, const gchar *name,
			 PgmRenderTextWeight weight,
			 PgmRenderTextSlant slant, gint size)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  g_free (gl1_font->text_font_name);
  gl1_font->text_font_name = g_strndup (name, strlen (name));
  gl1_font->text_weight = weight;
  gl1_font->text_slant = slant;
  gl1_font->text_size = size;

  gl1_font->text_coord[0].s = 0.0f; gl1_font->text_coord[0].t = 0.0f;
  gl1_font->text_coord[0].r = 0.0f; gl1_font->text_coord[0].q = 1.0f;
  gl1_font->text_coord[1].s = 0.0f; gl1_font->text_coord[1].t = 1.0f;
  gl1_font->text_coord[1].r = 0.0f; gl1_font->text_coord[1].q = 1.0f;
  gl1_font->text_coord[2].s = 1.0f; gl1_font->text_coord[2].t = 1.0f;
  gl1_font->text_coord[2].r = 0.0f; gl1_font->text_coord[2].q = 1.0f;
  gl1_font->text_coord[3].s = 1.0f; gl1_font->text_coord[3].t = 0.0f;
  gl1_font->text_coord[3].r = 0.0f; gl1_font->text_coord[3].q = 1.0f;
  
  generate_cairo_label (gl1_font);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_itext_get_font (PgmRenderTextIFace *text, PgmRenderTextWeight *weight,
			 PgmRenderTextSlant *slant, gint *size)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  g_return_val_if_fail (weight, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (slant, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (size, PGM_RENDER_ERROR_X);

  *weight = gl1_font->text_weight;
  *slant = gl1_font->text_slant;
  *size = gl1_font->text_size;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_itext_set_color (PgmRenderTextIFace *text, guchar r, guchar g,
			  guchar b, guchar a)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  gl1_font->text_color.r = (gfloat) r * INV_255;
  gl1_font->text_color.g = (gfloat) g * INV_255;
  gl1_font->text_color.b = (gfloat) b * INV_255;
  gl1_font->text_color.a = (gfloat) a * INV_255;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_itext_get_color (PgmRenderTextIFace *text, guchar *r, guchar *g,
			  guchar *b, guchar *a)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  g_return_val_if_fail (r, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (g, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (b, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (a, PGM_RENDER_ERROR_X);

  *r = (guchar) (gl1_font->text_color.r * 255);
  *g = (guchar) (gl1_font->text_color.g * 255);
  *b = (guchar) (gl1_font->text_color.b * 255);
  *a = (guchar) (gl1_font->text_color.a * 255);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_itext_get_size (PgmRenderTextIFace *text, gint *text_width,
			 gint *text_height)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (text);

  g_return_val_if_fail (text_width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (text_height, PGM_RENDER_ERROR_X);

  *text_width = gl1_font->text_text_extents.width + 2;
  *text_height = (gl1_font->text_font_extents.ascent +
                  gl1_font->text_font_extents.descent) + 2;

  return PGM_RENDER_ERROR_X;
}

static void
text_iface_init (PgmRenderTextIFaceClass *iface)
{
  iface->set_label = gl1_font_itext_set_label;
  iface->get_label = gl1_font_itext_get_label;
  iface->set_font = gl1_font_itext_set_font;
  iface->get_font = gl1_font_itext_get_font;
  iface->set_color = gl1_font_itext_set_color;
  iface->get_color = gl1_font_itext_get_color;
  iface->get_size = gl1_font_itext_get_size;
}

/* Pipe3d interface implementation */

static PgmRenderError
gl1_font_ipipe3d_set_translation (PgmRenderPipe3dIFace *obj, gfloat x,
				  gfloat y, gfloat z)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);
  PgmRenderDrawable *drawable = PGM_RENDER_DRAWABLE (obj);

  gl1_font->translation[0] = x;
  gl1_font->translation[1] = y;
  gl1_font->translation[2] = z;

  if (drawable->window && drawable->order == PGM_RENDER_DRAWABLE_Z_ORDERED)
    pgm_render_base_window_reorder_drawable (drawable->window, drawable, 0);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ipipe3d_get_translation (PgmRenderPipe3dIFace *obj, gfloat *x,
				  gfloat *y, gfloat *z)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (z, PGM_RENDER_ERROR_X);

  *x = gl1_font->translation[0];
  *y = gl1_font->translation[1];
  *z = gl1_font->translation[2];

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ipipe3d_set_rotation (PgmRenderPipe3dIFace *obj, gfloat x,
				  gfloat y, gfloat z)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  gl1_font->rotation[0] = x;
  gl1_font->rotation[1] = y;
  gl1_font->rotation[2] = z;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_font_ipipe3d_get_rotation (PgmRenderPipe3dIFace *obj, gfloat *x,
			       gfloat *y, gfloat *z)
{
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (z, PGM_RENDER_ERROR_X);

  *x = gl1_font->rotation[0];
  *y = gl1_font->rotation[1];
  *z = gl1_font->rotation[2];

  return PGM_RENDER_ERROR_X;
}

static void
pipe3d_iface_init (PgmRenderPipe3dIFaceClass *iface)
{
  iface->set_translation = gl1_font_ipipe3d_set_translation;
  iface->get_translation = gl1_font_ipipe3d_get_translation;
  iface->set_rotation = gl1_font_ipipe3d_set_rotation;
  iface->get_rotation = gl1_font_ipipe3d_get_rotation;
}

static PgmRenderError
pgm_render_gl1_font_draw (PgmRenderDrawable *drawable)
{
  PgmRenderGlProcAddressList *gl = PGM_RENDER_GL1_WINDOW
    (drawable->window)->backend->proc_address;
  PgmRenderGl1Font *gl1_font = PGM_RENDER_GL1_FONT (drawable);
  gfloat x_offset, y_offset;
  gint i;

  PGM_RENDER_DRAWABLE_LOCK (drawable);
  x_offset = (gl1_font->text_vertex[3].x + gl1_font->text_vertex[0].x) * 0.5f;
  y_offset = (gl1_font->text_vertex[1].y + gl1_font->text_vertex[0].y) * 0.5f;

  gl->load_identity ();
  gl->translate_f (gl1_font->translation[0] + x_offset,
                   gl1_font->translation[1] + y_offset,
                   gl1_font->translation[2]);
  gl->rotate_f (gl1_font->rotation[0], 1.0f, 0.0f, 0.0f);
  gl->rotate_f (gl1_font->rotation[1], 0.0f, 1.0f, 0.0f);
  gl->rotate_f (gl1_font->rotation[2], 0.0f, 0.0f, 1.0f);
  gl->translate_f (-x_offset, -y_offset, 0.0f);

  gl->enable (PGM_RENDER_GL_TEXTURE_2D);
  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, gl1_font->text_id);
  gl->begin (PGM_RENDER_GL_QUADS);
  gl->color_4f (gl1_font->text_color.r, gl1_font->text_color.g,
                gl1_font->text_color.b, gl1_font->text_color.a);
  for (i = 0; i < 4; i++)
    {
      gl->tex_coord_2f (gl1_font->text_coord[i].s, gl1_font->text_coord[i].t);
      gl->vertex_3f (gl1_font->text_vertex[i].x, gl1_font->text_vertex[i].y,
                     gl1_font->text_vertex[i].z);
    }
  gl->end ();
  PGM_RENDER_DRAWABLE_UNLOCK (drawable);

  return PGM_RENDER_ERROR_X;
}

void
pgm_render_gl1_font_set_scale (PgmRenderGl1Font *font, gfloat sx, gfloat sy)
{
/*   font->sx *= sx / font->window_sx; */
/*   font->sy *= sy / font->window_sy; */
/*   font->window_sx = sx; */
/*   font->window_sy = sy; */

  font->window_sx = sx;
  font->window_sy = sy;
  font->sx = 1.0f;
  font->sy = 1.0f;
  generate_cairo_label (font);
}

void
pgm_render_gl1_font_regenerate (PgmRenderGl1Font *font)
{
  return;
}
